@@ -300,7 +300,7 @@ int CmdExec::Done()
 {
    Enter();
    bool done = (feeder==0 && Idle())
-      || (auto_terminate_in_bg && NumberOfJobs()==0 && !in_foreground_pgrp());
+      || (auto_terminate_in_bg && NumberOfChildrenJobs()==0 && !in_foreground_pgrp());
    Leave();
    return done;
 }
@@ -822,12 +822,12 @@ void CmdExec::init(LocalDirectory *c)
    Reconfig();
 }
 CmdExec::CmdExec(FileAccess *s,LocalDirectory *c)
-   : SessionJob(s?s:new DummyProto)
+   : SessionJob(s?s:new DummyProto), parent_exec(0)
 {
    init(c);
 }
-CmdExec::CmdExec(const CmdExec *parent)
-   : SessionJob(parent->session->Clone())
+CmdExec::CmdExec(CmdExec *parent)
+   : SessionJob(parent->session->Clone()), parent_exec(parent)
 {
    init(parent->cwd->Clone());
 }
