@@ -43,7 +43,7 @@
 # include "module.h"
 #endif
 
-FileAccess *FileAccess::chain=0;
+xlist_head<FileAccess> FileAccess::all_fa;
 const FileAccessRef FileAccessRef::null;
 
 void FileAccess::Init()
@@ -54,12 +54,10 @@ void FileAccess::Init()
 
    default_cwd="~";
    cwd.Set(default_cwd,false,0);
-   new_cwd=0;
    limit=FILE_END;
    real_pos=UNKNOWN_POS;
    pos=0;
    mode=CLOSED;
-   try_time=0;
    retries=0;
    opt_date=0;
    opt_size=0;
@@ -79,11 +77,11 @@ void FileAccess::Init()
 
    priority=last_priority=0;
 
-   next=chain;
-   chain=this;
+   all_fa.add(all_fa_node);
 }
 
 FileAccess::FileAccess(const FileAccess *fa)
+   : all_fa_node(this)
 {
    Init();
 
@@ -99,7 +97,7 @@ FileAccess::FileAccess(const FileAccess *fa)
 
 FileAccess::~FileAccess()
 {
-   ListDel(FileAccess,chain,this,next);
+   all_fa_node.remove();
 }
 
 void  FileAccess::Open(const char *fn,int mode,off_t offs)
@@ -188,12 +186,13 @@ void FileAccess::Close()
    file.set(0);
    file_url.set(0);
    file1.set(0);
-   delete new_cwd; new_cwd=0;
+   new_cwd=0;
    mode=CLOSED;
    opt_date=0;
    opt_size=0;
    fileset_for_info=0;
    retries=0;
+   max_retries=0;
    entity_size=NO_SIZE;
    entity_date=NO_DATE;
    ascii=false;
@@ -285,7 +284,7 @@ const char *FileAccess::GetFileURL(const char *f,int flags) const
       if(!f || (f[0]!='/' && f[0]!='~'))
 	 f=dir_file(cwd.path?cwd.path.get():"~",f);
       u.path.set(f);
-      if(is_dir && url::dir_needs_trailing_slash(proto))
+      if(is_dir && url::dir_needs_trailing_slash(proto) && u.path.last_char()!='/')
 	 u.path.append('/');
    }
    return url.set_allocated(u.Combine(home));
@@ -314,8 +313,7 @@ void FileAccess::Login(const char *user1,const char *pass1)
 
    if(user && pass==0)
    {
-      FileAccess *o;
-      for(o=chain; o!=0; o=o->next)
+      xlist_for_each(FileAccess,all_fa,node,o)
       {
 	 pass.set(o->pass);
 	 if(SameSiteAs(o) && o->pass)
@@ -352,7 +350,7 @@ void FileAccess::SetPasswordGlobal(const char *p)
 {
    pass.set(p);
    xstring save_pass;
-   for(FileAccess *o=chain; o!=0; o=o->next)
+   xlist_for_each(FileAccess,all_fa,node,o)
    {
       if(o==this)
 	 continue;
@@ -539,7 +537,6 @@ void FileAccess::Chdir(const char *path,bool verify)
    else
    {
       cwd.Set(new_cwd);
-      delete new_cwd;
       new_cwd=0;
    }
 }
@@ -697,42 +694,12 @@ void SessionPool::ClearAll()
    }
 }
 
-bool FileAccess::NotSerious(int e)
-{
-   switch(e)
-   {
-   case(EPIPE):
-   case(ETIMEDOUT):
-#ifdef ECONNRESET
-   case(ECONNRESET):
-#endif
-   case(ECONNREFUSED):
-#ifdef EHOSTUNREACH
-   case(EHOSTUNREACH):
-#endif
-#ifdef EHOSTDOWN
-   case(EHOSTDOWN):
-#endif
-#ifdef ENETRESET
-   case(ENETRESET):
-#endif
-#ifdef ENETUNREACH
-   case(ENETUNREACH):
-#endif
-#ifdef ENETDOWN
-   case(ENETDOWN):
-#endif
-#ifdef ECONNABORTED
-   case(ECONNABORTED):
-#endif
-      return true;
-   }
-   return false;
-}
-
 void FileAccess::SetTryTime(time_t t)
 {
-   try_time=t;
+   if(t)
+      reconnect_timer.Reset(Time(t));
+   else
+      reconnect_timer.Stop();
 }
 
 bool FileAccess::IsBetterThan(const FileAccess *fa) const
@@ -746,7 +713,6 @@ const char *FileAccess::CurrentStatus() { return ""; }
 int FileAccess::Buffered() { return 0; }
 bool FileAccess::IOReady() { return IsOpen(); }
 int FileAccess::IsConnected() const { return 0; }
-void FileAccess::Disconnect() {}
 void FileAccess::UseCache(bool) {}
 bool FileAccess::NeedSizeDateBeforehand() { return false; }
 void FileAccess::Cleanup() {}
@@ -757,7 +723,7 @@ DirList *FileAccess::MakeDirList(ArgV *a) { delete a; return 0; }
 
 void FileAccess::CleanupAll()
 {
-   for(FileAccess *o=chain; o!=0; o=o->next)
+   xlist_for_each(FileAccess,all_fa,node,o)
    {
       Enter(o);
       o->CleanupThis();
@@ -768,10 +734,10 @@ void FileAccess::CleanupAll()
 FileAccess *FileAccess::NextSameSite(FA *scan) const
 {
    if(scan==0)
-      scan=chain;
+      scan=all_fa.first_obj();
    else
-      scan=scan->next;
-   for( ; scan; scan=scan->next)
+      scan=scan->all_fa_node.next_obj();
+   for( ; scan; scan=scan->all_fa_node.next_obj())
       if(scan!=this && SameSiteAs(scan))
 	 return scan;
    return 0;
@@ -800,7 +766,7 @@ FileAccess *FileAccess::New(const char *proto,const char *host,const char *port)
       FA *n_session=Protocol::NewSession(n_proto);
       if(n_session)
       {
-	 delete session;
+	 Delete(session);
 	 session=n_session;
 	 session->SetVisualProto(proto);
       }
@@ -830,22 +796,18 @@ FileAccess *FileAccess::New(const ParsedURL *u,bool dummy)
 }
 
 // FileAccess::Protocol implementation
-FileAccess::Protocol *FileAccess::Protocol::chain=0;
+xmap<FileAccess::Protocol*> FileAccess::Protocol::proto_by_name;
 
 FileAccess::Protocol::Protocol(const char *proto, SessionCreator *creator)
 {
    this->proto=proto;
    this->New=creator;
-   this->next=chain;
-   chain=this;
+   proto_by_name.add(proto,this);
 }
 
 FileAccess::Protocol *FileAccess::Protocol::FindProto(const char *proto)
 {
-   for(Protocol *scan=chain; scan; scan=scan->next)
-      if(!strcasecmp(scan->proto,proto))
-	 return scan;
-   return 0;
+   return proto_by_name.lookup(proto);
 }
 
 FileAccess *FileAccess::Protocol::NewSession(const char *proto)
@@ -962,7 +924,7 @@ void FileAccess::Path::Change(const char *new_path,bool new_is_file,const char *
 	    new_url_path.set("/~");
       }
       if(new_url_path.last_char()!='/')
-	 new_url_path.append("/");
+	 new_url_path.append('/');
       if(new_path[0]=='/' || new_path[0]=='~' || new_device_prefix_len!=0)
       {
 	 bool have_slash=((new_path_enc?new_path_enc:new_path)[0]=='/');
