@@ -181,8 +181,8 @@ static int sort_dirs(const void *s1, const void *s2)
 {
    const FileInfo *p1=(*files_cmp)[*static_cast<const int*>(s1)];
    const FileInfo *p2=(*files_cmp)[*static_cast<const int*>(s2)];
-   if(p1->filetype == FileInfo::DIRECTORY && !p2->filetype == FileInfo::DIRECTORY) return -rev_cmp;
-   if(!p1->filetype == FileInfo::DIRECTORY && p2->filetype == FileInfo::DIRECTORY) return rev_cmp;
+   if((p1->filetype == FileInfo::DIRECTORY) && !(p2->filetype == FileInfo::DIRECTORY)) return -rev_cmp;
+   if(!(p1->filetype == FileInfo::DIRECTORY) && (p2->filetype == FileInfo::DIRECTORY)) return rev_cmp;
    return 0;
 }
 
@@ -232,6 +232,21 @@ void FileSet::Sort(sort_e newsort, bool casefold, bool reverse)
    }
 }
 
+// reverse current sort order
+void FileSet::ReverseSort()
+{
+   if(!sorted) {
+      Sort(BYNAME,false,true);
+      return;
+   }
+   int i=0;
+   int j=sorted.length()-1;
+   while(i<j) {
+      sorted[i]=replace_value(sorted[j],sorted[i]);
+      ++i,--j;
+   }
+}
+
 /* Remove the current sort, allowing new entries to be added.
  * (Nothing uses this ... */
 void FileSet::Unsort()
