@@ -98,6 +98,15 @@ int Fish::Do()
       timeout_timer.Reset(send_buf->EventTime());
    if(recv_buf)
       timeout_timer.Reset(recv_buf->EventTime());
+   if(pty_send_buf)
+      timeout_timer.Reset(pty_send_buf->EventTime());
+   if(pty_recv_buf)
+      timeout_timer.Reset(pty_recv_buf->EventTime());
+
+   // check for timeout only if there should be connection activity.
+   if(state!=DISCONNECTED && (state!=CONNECTED || !RespQueueIsEmpty())
+   && mode!=CLOSED && CheckTimeout())
+      return MOVED;
 
    if((state==FILE_RECV || state==FILE_SEND)
    && rate_limit==0)
@@ -275,14 +284,6 @@ int Fish::Do()
    case DONE:
       break;
    }
-   if(m==MOVED)
-      return MOVED;
-   if(send_buf)
-      timeout_timer.Reset(send_buf->EventTime());
-   if(recv_buf)
-      timeout_timer.Reset(recv_buf->EventTime());
-   if(CheckTimeout())
-      return MOVED;
    return m;
 }
 
@@ -301,9 +302,9 @@ void Fish::MoveConnectionHere(Fish *o)
    ResumeInternal();
 }
 
-void Fish::Disconnect()
+void Fish::DisconnectLL()
 {
-   super::Disconnect();
+   super::DisconnectLL();
    EmptyRespQueue();
    EmptyPathQueue();
    state=DISCONNECTED;
@@ -795,7 +796,7 @@ const char *memstr(const char *mem,size_t len,const char *str)
    return 0;
 }
 
-int Fish::Read(void *buf,int size)
+int Fish::Read(Buffer *buf,int size)
 {
    if(Error())
       return error_code;
@@ -871,8 +872,7 @@ int Fish::Read(void *buf,int size)
       }
       if(size>size1)
 	 size=size1;
-      memcpy(buf,buf1,size);
-      recv_buf->Skip(size);
+      buf->MoveDataHere(recv_buf,size);
       pos+=size;
       real_pos+=size;
       rate_limit->BytesGot(size);
