@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2013 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2014 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -52,13 +52,65 @@ CDECL char *strptime(const char *buf, const char *format, struct tm *tm);
 #define HTTP_DEFAULT_PROXY_PORT	 "3128"
 #define HTTPS_DEFAULT_PORT	 "443"
 
+enum {
+   H_Continue=100,
+   H_Switching_Protocols=101,
+   H_Processing=102,
+   H_Ok=200,
+   H_Created=201,
+   H_Accepted=202,
+   H_Non_Authoritative_Information=203,
+   H_No_Content=204,
+   H_Reset_Content=205,
+   H_Partial_Content=206,
+   H_Multi_Status=207,
+   H_Moved_Permanently=301,
+   H_Found=302,
+   H_See_Other=303,
+   H_Not_Modified=304,
+   H_Use_Proxy=305,
+   H_Temporary_Redirect=307,
+   H_Bad_Request=400,
+   H_Unauthorized=401,
+   H_Payment_Required=402,
+   H_Forbidden=403,
+   H_Not_Found=404,
+   H_Method_Not_Allowed=405,
+   H_Not_Acceptable=406,
+   H_Proxy_Authentication_Required=407,
+   H_Request_Timeout=408,
+   H_Conflict=409,
+   H_Gone=410,
+   H_Length_Required=411,
+   H_Precondition_Failed=412,
+   H_Request_Entity_Too_Large=413,
+   H_Request_URI_Too_Long=414,
+   H_Unsupported_Media_Type=415,
+   H_Requested_Range_Not_Satisfiable=416,
+   H_Expectation_Failed=417,
+   H_Unprocessable_Entity=422,
+   H_Locked=423,
+   H_Failed_Dependency=424,
+   H_Too_Many_Requests=429,
+   H_Internal_Server_Error=500,
+   H_Not_Implemented=501,
+   H_Bad_Gateway=502,
+   H_Service_Unavailable=503,
+   H_Gateway_Timeout=504,
+   H_HTTP_Version_Not_Supported=505,
+   H_Insufficient_Storage=507,
+};
+
 /* Some status code validation macros: */
-#define H_20X(x)        (((x) >= 200) && ((x) < 300))
-#define H_PARTIAL(x)    ((x) == 206)
-#define H_REDIRECTED(x) (((x) == 301) || ((x) == 302) || ((x) == 303) || ((x) == 307))
-#define H_EMPTY(x)	(((x) == 204) || ((x) == 205))
-#define H_CONTINUE(x)	((x) == 100 || (x) == 102)
-#define H_REQUESTED_RANGE_NOT_SATISFIABLE(x) ((x) == 416)
+#define H_2XX(x)        (((x) >= 200) && ((x) <= 299))
+#define H_5XX(x)        (((x) >= 500) && ((x) <= 599))
+#define H_PARTIAL(x)    ((x) == H_Partial_Content)
+#define H_REDIRECTED(x) (((x) == H_Moved_Permanently) || ((x) == H_Found) || ((x) == H_See_Other) || ((x) == H_Temporary_Redirect))
+#define H_EMPTY(x)	(((x) == H_No_Content) || ((x) == H_Reset_Content))
+#define H_CONTINUE(x)	((x) == H_Continue || (x) == H_Processing)
+#define H_REQUESTED_RANGE_NOT_SATISFIABLE(x) ((x) == H_Requested_Range_Not_Satisfiable)
+#define H_TRANSIENT(x)	((x)==H_Request_Timeout || (x)==H_Bad_Gateway || (x)==H_Service_Unavailable || (x)==H_Gateway_Timeout)
+#define H_UNSUPPORTED(x) ((x)==H_Bad_Request || (x)==H_Not_Implemented)
 
 #ifndef EINPROGRESS
 #define EINPROGRESS -1
@@ -158,7 +210,7 @@ void Http::MoveConnectionHere(Http *o)
    ResumeInternal();
 }
 
-void Http::Disconnect()
+void Http::DisconnectLL()
 {
    rate_limit=0;
    if(conn)
@@ -230,8 +282,8 @@ void Http::Close()
    else
    {
    disconnect:
-      try_time=0;
       Disconnect();
+      DontSleep();
    }
    array_send=0;
    no_cache_this=false;
@@ -316,6 +368,9 @@ void Http::SendMethod(const char *method,const char *efile)
       accept=Query("accept-charset",hostname);
       if(accept && accept[0])
 	 Send("Accept-Charset: %s\r\n",accept);
+      accept=Query("accept-encoding",hostname);
+      if(accept && accept[0])
+	 Send("Accept-Encoding: %s\r\n",accept);
 
       const char *referer=Query("referer",hostname);
       const char *slash="";
@@ -649,10 +704,14 @@ void Http::SendRequest(const char *connection,const char *f)
       break;
 
    case(REMOVE):
-   case(REMOVE_DIR):
       SendMethod("DELETE",efile);
-      if(mode==REMOVE)
-	 Send("Depth: 0\r\n"); // deny directory removal
+      Send("Depth: 0\r\n"); // deny directory removal
+      break;
+
+   case(REMOVE_DIR):
+      if(efile.last_char()!='/')
+	 efile.append('/');
+      SendMethod("DELETE",efile);
       break;
 
    case ARRAY_INFO:
@@ -736,6 +795,41 @@ int Http::SendArrayInfoRequest()
    return req_count;
 }
 
+void Http::ProceedArrayInfo()
+{
+   for(;;)
+   {
+      // skip to next needed file
+      FileInfo *fi=fileset_for_info->next();
+      if(!fi || fi->need)
+	 break;
+   }
+   if(!fileset_for_info->curr())
+   {
+      LogNote(10,"that was the last file info");
+      // received all requested info.
+      state=DONE;
+      return;
+   }
+   // we can avoid reconnection if server supports it.
+   if(keep_alive && (keep_alive_max>1 || keep_alive_max==-1)
+   && (use_head || use_propfind_now))
+   {
+      // we'll have to receive next header, unset the status
+      status.set(0);
+      status_code=0;
+
+      state=CONNECTED;
+      SendArrayInfoRequest();
+      state=RECEIVING_HEADER;
+   }
+   else
+   {
+      Disconnect();
+      DontSleep();
+   }
+}
+
 static const char *extract_quoted_header_value(const char *value)
 {
    static xstring value1;
@@ -763,20 +857,20 @@ static const char *extract_quoted_header_value(const char *value)
 
 void Http::HandleHeaderLine(const char *name,const char *value)
 {
-   if(!strcasecmp(name,"Content-length"))
-   {
+   // use a perfect hash
+#define hh(L,C) ((L)+(C)*3)
+#define hhc(S,C) hh(sizeof((S))-1,(C))
+#define case_hh(S,C) case hhc((S),(C)): if(strcasecmp(name,(S))) break;
+   switch(hh(strlen(name),c_toupper(name[0]))) {
+   case_hh("Content-Length",'C') {
       long long bs=0;
       if(1!=sscanf(value,"%lld",&bs))
 	 return;
       if(bs<0) // try to workaround broken servers
 	 bs+=0x100000000LL;
       body_size=bs;
-      if(pos==0 && mode!=STORE && mode!=MAKE_DIR && !inflate)
-	 entity_size=body_size;
-      if(pos==0 && opt_size && H_20X(status_code) && !inflate)
-	 *opt_size=body_size;
-
-      if(mode==ARRAY_INFO && H_20X(status_code))
+      if(mode==ARRAY_INFO && H_2XX(status_code)
+      && xstrcmp(last_method,"PROPFIND"))
       {
 	 FileInfo *fi=fileset_for_info->curr();
 	 fi->SetSize(body_size);
@@ -784,8 +878,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       }
       return;
    }
-   if(!strcasecmp(name,"Content-range"))
-   {
+   case_hh("Content-Range",'C') {
       long long first,last,fsize;
       if(H_REQUESTED_RANGE_NOT_SATISFIABLE(status_code))
       {
@@ -804,13 +897,12 @@ void Http::HandleHeaderLine(const char *name,const char *value)
 	 body_size=last-first+1;
       if(mode!=STORE && mode!=MAKE_DIR)
 	 entity_size=fsize;
-      if(opt_size && H_20X(status_code))
+      if(opt_size && H_2XX(status_code))
 	 *opt_size=fsize;
       return;
    }
-   if(!strcasecmp(name,"Last-Modified"))
-   {
-      if(!H_20X(status_code))
+   case_hh("Last-Modified",'L') {
+      if(!H_2XX(status_code))
 	 return;
 
       time_t t=Http::atotm(value);
@@ -820,7 +912,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       if(opt_date)
 	 *opt_date=t;
 
-      if(mode==ARRAY_INFO)
+      if(mode==ARRAY_INFO && !propfind)
       {
 	 FileInfo *fi=fileset_for_info->curr();
 	 fi->SetDate(t,0);
@@ -828,19 +920,16 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       }
       return;
    }
-   if(!strcasecmp(name,"Location"))
-   {
+   case_hh("Location",'L')
       location.set(value);
       return;
-   }
-   if(!strcasecmp(name,"Retry-After"))
-   {
+
+   case_hh("Retry-After",'R')
       retry_after=0;
       sscanf(value,"%ld",&retry_after);
       return;
-   }
-   if(!strcasecmp(name,"Keep-Alive"))
-   {
+
+   case_hh("Keep-Alive",'K') {
       keep_alive=true;
       const char *m=strstr(value,"max=");
       if(m) {
@@ -850,20 +939,20 @@ void Http::HandleHeaderLine(const char *name,const char *value)
 	 keep_alive_max=100;
       return;
    }
-   if(!strcasecmp(name,"Connection")
-   || !strcasecmp(name,"Proxy-Connection"))
-   {
+   case_hh("Proxy-Connection",'P')
+      goto case_Connection;
+   case_hh("Connection",'C')
+   case_Connection:
       if(!strcasecmp(value,"keep-alive"))
 	 keep_alive=true;
       else if(!strcasecmp(value,"close"))
 	 keep_alive=false;
       return;
-   }
-   if(!strcasecmp(name,"Transfer-Encoding"))
-   {
+
+   case_hh("Transfer-Encoding",'T')
       if(!strcasecmp(value,"identity"))
 	 return;
-      if(!strcasecmp(value,"chunked"));
+      if(!strcasecmp(value,"chunked"))
       {
 	 chunked=true;
 	 chunk_size=CHUNK_SIZE_UNKNOWN;	  // expecting first chunk
@@ -871,35 +960,24 @@ void Http::HandleHeaderLine(const char *name,const char *value)
 	 chunked_trailer=false;
       }
       return;
-   }
-   if(!strcasecmp(name,"Content-Encoding"))
-   {
-      if(!strcasecmp(value,"gzip")) {
-	 // inflated size if unknown beforehand
-	 entity_size=NO_SIZE;
-	 if(opt_size)
-	    *opt_size=NO_SIZE;
-	 // start the inflation
-	 inflate=new DirectedBuffer(DirectedBuffer::GET);
-	 inflate->SetTranslator(new DataInflator());
-      }
-   }
-   if(!strcasecmp(name,"Accept-Ranges"))
-   {
+
+   case_hh("Content-Encoding",'C')
+      content_encoding.set(value);
+      return;
+
+   case_hh("Accept-Ranges",'A')
       if(!strcasecmp(value,"none"))
 	 no_ranges=true;
       if(strstr(value,"bytes"))
 	 seen_ranges_bytes=true;
       return;
-   }
-   if(!strcasecmp(name,"Set-Cookie"))
-   {
+
+   case_hh("Set-Cookie",'S')
       if(!hftp && QueryBool("set-cookies",hostname))
 	 SetCookie(value);
       return;
-   }
-   if(!strcasecmp(name,"Content-Disposition"))
-   {
+
+   case_hh("Content-Disposition",'C') {
       const char *filename=strstr(value,"filename=");
       if(!filename)
 	 return;
@@ -907,8 +985,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       SetSuggestedFileName(filename);
       return;
    }
-   if(!strcasecmp(name,"Content-Type"))
-   {
+   case_hh("Content-Type",'C') {
       entity_content_type.set(value);
       const char *cs=strstr(value,"charset=");
       if(cs)
@@ -916,7 +993,12 @@ void Http::HandleHeaderLine(const char *name,const char *value)
 	 cs=extract_quoted_header_value(cs+8);
 	 entity_charset.set(cs);
       }
+      return;
    }
+   default:
+      break;
+   }
+   LogNote(10,"unhandled header line `%s'",name);
 }
 
 static
@@ -971,6 +1053,7 @@ int Http::Do()
 {
    int m=STALL;
    int res;
+   const char *error;
    const char *buf;
    int len;
 
@@ -983,40 +1066,56 @@ int Http::Do()
    }
 
    if(home.path==0)
-      home.Set(default_cwd);
-   ExpandTildeInCWD();
+      set_home(default_cwd);
 
    if(Error())
       return m;
 
-   if(propfind && mode==CHANGE_DIR)
+   if(propfind)
    {
+      if(propfind->Error())
+      {
+	 propfind=0;
+	 if(mode==CHANGE_DIR)
+	 {
+	    SetError(NO_FILE,propfind->ErrorText());
+	    return MOVED;
+	 }
+	 if(propfind->ErrorFatal())
+	    fileset_for_info->next();
+	 Disconnect();
+	 return MOVED;
+      }
       if(propfind->Eof())
       {
 	 const char *b;
 	 int len;
 	 propfind->Get(&b,&len);
 	 FileSet *fs=HttpListInfo::ParseProps(b,len,GetCwd());
+	 propfind=0;
 	 if(fs)
 	 {
+	    if(mode==ARRAY_INFO)
+	       fileset_for_info->Merge(fs);
 	    FileInfo *fi=fs->curr();
 	    if(fi && fi->Has(fi->TYPE))
 	    {
-	       new_cwd->is_file=(fi->filetype!=fi->DIRECTORY);
+	       if(mode==CHANGE_DIR)
+		  new_cwd->is_file=(fi->filetype!=fi->DIRECTORY);
 	       fi->MakeLongName();
 	       LogNote(9,"%s",fi->longname.get());
 	    }
 	 }
-	 cwd.Set(new_cwd);
-	 cache->SetDirectory(this, "", !cwd.is_file);
+	 m=MOVED;
 	 state=DONE;
-	 propfind=0;
-	 return MOVED;
-      }
-      if(propfind->Error()) {
-	 SetError(NO_FILE,propfind->ErrorText());
-	 propfind=0;
-	 return MOVED;
+	 if(mode==CHANGE_DIR)
+	 {
+	    cwd.Set(new_cwd);
+	    cache->SetDirectory(this, "", !cwd.is_file);
+	    return m;
+	 }
+	 else if(mode==ARRAY_INFO)
+	    ProceedArrayInfo();
       }
    }
 
@@ -1176,11 +1275,12 @@ int Http::Do()
       timeout_timer.Reset();
 
    case CONNECTING:
-      res=Poll(conn->sock,POLLOUT);
+      res=Poll(conn->sock,POLLOUT,&error);
       if(res==-1)
       {
+	 LogError(0,_("Socket error (%s) - reconnecting"),error);
+	 Disconnect(error);
 	 NextPeer();
-	 Disconnect();
 	 return MOVED;
       }
       if(!(res&POLLOUT))
@@ -1239,6 +1339,7 @@ int Http::Do()
 	 SetError(NOT_SUPP);
 	 return MOVED;
       }
+      ExpandTildeInCWD();
       if(mode==STORE && pos>0 && entity_size>=0 && pos>=entity_size)
       {
 	 state=DONE;
@@ -1266,7 +1367,7 @@ int Http::Do()
    case RECEIVING_HEADER:
       if(conn->send_buf->Error() || conn->recv_buf->Error())
       {
-	 if((mode==STORE || special) && status_code && !H_20X(status_code))
+	 if((mode==STORE || special) && status_code && !H_2XX(status_code))
 	    goto pre_RECEIVING_BODY;   // assume error.
       handle_buf_error:
 	 if(conn->send_buf->Error())
@@ -1312,7 +1413,7 @@ int Http::Do()
 	       conn->recv_buf->Skip(eol_size);
 	       if(tunnel_state==TUNNEL_WAITING)
 	       {
-		  if(H_20X(status_code))
+		  if(H_2XX(status_code))
 		  {
 #if USE_SSL
 		     if(https)
@@ -1328,7 +1429,12 @@ int Http::Do()
 	       {
 		  chunked_trailer=false;
 		  chunked=false;
-		  state=DONE;
+		  if(propfind)
+		     propfind->PutEOF();
+		  if(mode==ARRAY_INFO)
+		     state=CONNECTED;
+		  else
+		     state=DONE;
 		  return MOVED;
 	       }
 	       if(H_CONTINUE(status_code))
@@ -1339,58 +1445,34 @@ int Http::Do()
 	       }
 	       if(mode==ARRAY_INFO)
 	       {
-		  if((status_code==400 || status_code==501)
-		  && !xstrcmp(last_method,"PROPFIND"))
+		  if(!xstrcmp(last_method,"PROPFIND"))
 		  {
-		     ResMgr::Set("http:use-propfind",hostname,"no");
-		     use_propfind_now=false;
-		     try_time=0;
-		     Disconnect();
-		     return MOVED;
-		  }
-		  // we'll have to receive next header
-		  status.set(0);
-		  status_code=0;
-		  for(;;)
-		  {
-		     // skip to next needed file
-		     FileInfo *fi=fileset_for_info->next();
-		     if(!fi || fi->need)
-			break;
-		  }
-		  if(!fileset_for_info->curr())
-		  {
-		     // received all requested info.
-		     state=DONE;
-		     return MOVED;
-		  }
-		  // we can avoid reconnection if server supports it.
-		  if(keep_alive && (keep_alive_max>1 || keep_alive_max==-1) && use_head)
-		  {
-		     state=CONNECTED;
-		     if(SendArrayInfoRequest()==0)
-			state=DONE;
-		  }
-		  else
-		  {
-		     Disconnect();
-		     try_time=0;
+		     if(H_UNSUPPORTED(status_code))
+		     {
+			ResMgr::Set("http:use-propfind",hostname,"no");
+			use_propfind_now=false;
+			Disconnect();
+			DontSleep();
+			return MOVED;
+		     }
+		     goto pre_RECEIVING_BODY;
 		  }
+		  ProceedArrayInfo();
 		  return MOVED;
 	       }
 	       else if(mode==STORE || mode==MAKE_DIR)
 	       {
-		  if((sent_eot || pos==entity_size) && H_20X(status_code))
+		  if((sent_eot || pos==entity_size) && H_2XX(status_code))
 		  {
 		     state=DONE;
 		     Disconnect();
 		     state=DONE;
 		     return MOVED;
 		  }
-		  if(H_20X(status_code))
+		  if(H_2XX(status_code))
 		  {
 		     // should never happen
-		     LogError(0,"Success, but did nothing??");
+		     LogError(0,"Unexpected success, the server did not accept full request body");
 		     Disconnect();
 		     return MOVED;
 		  }
@@ -1416,7 +1498,7 @@ int Http::Do()
 	       {
 		  // simple 0.9 ?
 		  proto_version=0x09;
-		  status_code=200;
+		  status_code=H_Ok;
 		  LogError(0,_("Could not parse HTTP status line"));
 		  if(mode==STORE)
 		  {
@@ -1434,14 +1516,14 @@ int Http::Do()
 	       if(proto_version>=0x11)
 		  keep_alive=true;
 
-	       if(!H_20X(status_code))
+	       if(!H_2XX(status_code))
 	       {
 		  if(H_CONTINUE(status_code))
 		     return MOVED;
 
-		  if(status_code/100==5) // server failed, try another
+		  if(H_5XX(status_code)) // server failed, try another
 		     NextPeer();
-		  if(status_code==504) // Gateway Timeout
+		  if(status_code==H_Gateway_Timeout)
 		  {
 		     const char *cc=Query("cache-control");
 		     if(cc && strstr(cc,"only-if-cached"))
@@ -1451,23 +1533,20 @@ int Http::Do()
 			   SetError(NO_FILE,_("Object is not cached and http:cache-control has only-if-cached"));
 			   return MOVED;
 			}
-			status_code=406;  // so that no retry will be attempted
+			status_code=H_Not_Acceptable; // so that no retry will be attempted
 		     }
 		  }
 		  // check for retriable codes
-		  if(status_code==408  // Request Timeout
-		  || status_code==502  // Bad Gateway
-		  || status_code==503  // Service Unavailable
-		  || status_code==504) // Gateway Timeout
+		  if(H_TRANSIENT(status_code))
 		  {
 		     Disconnect();
 		     return MOVED;
 		  }
-		  if(status_code==429) // Too Many Requests
+		  if(status_code==H_Too_Many_Requests)
 		  {
 		     Disconnect();
 		     if(retry_after)
-			try_time = time_t(now)+retry_after-long(reconnect_interval_current);
+			reconnect_timer.StopDelayed(retry_after);
 		     return MOVED;
 		  }
 
@@ -1518,7 +1597,7 @@ int Http::Do()
 	 return MOVED;
       }
 
-      if(!H_20X(status_code))
+      if(!H_2XX(status_code))
       {
 	 xstring err;
 	 int code=NO_FILE;
@@ -1566,11 +1645,9 @@ int Http::Do()
 	 else
 	 {
 	    const char *closure=file;
-	    if(status_code==400  // Bad request
-	    || status_code==405  // Method Not Allowed
-	    || status_code==501) // Not Implemented
+	    if(H_UNSUPPORTED(status_code) || status_code==H_Method_Not_Allowed)
 	    {
-	       if(status_code==400 || status_code==501)
+	       if(H_UNSUPPORTED(status_code))
 	       {
 		  if(!xstrcmp(last_method,"PROPFIND"))
 		     ResMgr::Set("http:use-propfind",hostname,"no");
@@ -1580,8 +1657,8 @@ int Http::Do()
 	       if(mode==CHANGE_DIR && !xstrcmp(last_method,"PROPFIND"))
 	       {
 		  use_propfind_now=false;
-		  try_time=0;
 		  Disconnect();
+		  DontSleep();
 		  return MOVED;
 	       }
 	       code=NOT_SUPP;
@@ -1595,20 +1672,47 @@ int Http::Do()
 	 }
 	 state=RECEIVING_BODY;
 	 LogErrorText();
-	 SetError(code,err);
+	 if(mode==ARRAY_INFO)
+	 {
+	    if(!H_TRANSIENT(status_code))
+	       fileset_for_info->next();
+	    Disconnect();
+	    DontSleep();
+	 }
+	 else
+	    SetError(code,err);
 	 return MOVED;
       }
-      if(mode==CHANGE_DIR)
+      if(!xstrcmp(last_method,"PROPFIND")
+      && (mode==ARRAY_INFO || mode==CHANGE_DIR))
+	 propfind=new IOBufferFileAccess(this);
+
+      if(mode==CHANGE_DIR && !propfind)
       {
-	 if(!xstrcmp(last_method,"PROPFIND"))
-	    propfind=new IOBufferFileAccess(this);
-	 else
-	 {
-	    cwd.Set(new_cwd);
-	    cache->SetDirectory(this, "", !cwd.is_file);
-	    state=DONE;
-	    return MOVED;
-	 }
+	 cwd.Set(new_cwd);
+	 cache->SetDirectory(this, "", !cwd.is_file);
+	 state=DONE;
+	 return MOVED;
+      }
+
+      // Many servers send application/x-gzip with x-gzip encoding,
+      // don't decode in such a case.
+      if(CompressedContentEncoding() && !CompressedContentType()
+      && QueryBool("decode",hostname)) {
+	 // inflated size is unknown beforehand
+	 entity_size=NO_SIZE;
+	 if(opt_size)
+	    *opt_size=NO_SIZE;
+	 // start the inflation
+	 inflate=new DirectedBuffer(DirectedBuffer::GET);
+	 inflate->SetTranslator(new DataInflator());
+      }
+      // sometimes it's possible to derive entity size from body size.
+      if(entity_size==NO_SIZE && body_size!=NO_SIZE
+      && pos==0 && mode!=STORE && mode!=MAKE_DIR && !inflate) {
+	 entity_size=body_size;
+	 if(opt_size && H_2XX(status_code))
+	    *opt_size=body_size;
       }
 
       LogNote(9,_("Receiving body..."));
@@ -1618,7 +1722,7 @@ int Http::Do()
 	 if(mode!=STORE && mode!=MAKE_DIR && body_size>=0)
 	 {
 	    entity_size=body_size;
-	    if(opt_size && H_20X(status_code))
+	    if(opt_size && H_2XX(status_code))
 	       *opt_size=entity_size;
 	 }
 	 real_pos=0;
@@ -1632,7 +1736,7 @@ int Http::Do()
       if(conn->recv_buf->Size()>=rate_limit->BytesAllowedToGet())
       {
 	 conn->recv_buf->Suspend();
-	 Timeout(1000);
+	 TimeoutS(1);
       }
       else if(conn->recv_buf->Size()>=max_buf)
       {
@@ -1706,7 +1810,7 @@ void Http::ResumeInternal()
    super::ResumeInternal();
 }
 
-int Http::Read(void *buf,int size)
+int Http::Read(Buffer *buf,int size)
 {
    if(Error())
       return error_code;
@@ -1730,20 +1834,26 @@ int Http::Read(void *buf,int size)
 }
 void Http::_Skip(int to_skip)
 {
-   if(inflate) {
+   if(inflate)
       inflate->Skip(to_skip);
-   } else {
+   else
       conn->recv_buf->Skip(to_skip);
+   _UpdatePos(to_skip);
+}
+void Http::_UpdatePos(int to_skip)
+{
+   if(!inflate) {
       if(chunked)
 	 chunk_pos+=to_skip;
       bytes_received+=to_skip;
    }
    real_pos+=to_skip;
 }
-int Http::_Read(void *buf,int size)
+int Http::_Read(Buffer *buf,int size)
 {
    const char *buf1;
    int size1;
+   Buffer *src_buf=conn->recv_buf.get_non_const();
 get_again:
    if(conn->recv_buf->Size()==0 && conn->recv_buf->Error())
    {
@@ -1857,6 +1967,7 @@ get_again:
 	 bytes_received+=size1;
       }
       inflate->Get(&buf1,&size1);
+      src_buf=inflate.get_non_const();
    } else {
       if(size1>bytes_allowed)
 	 size1=bytes_allowed;
@@ -1875,8 +1986,8 @@ get_again:
    }
    if(size>size1)
       size=size1;
-   memcpy(buf,buf1,size);
-   _Skip(size);
+   buf->MoveDataHere(src_buf,size);
+   _UpdatePos(size);
    return size;
 }
 
@@ -2198,7 +2309,7 @@ void Http::CookieMerge(xstring &all,const char *cookie_c)
 
 void Http::MakeCookie(xstring &all_cookies,const char *hostname,const char *efile)
 {
-   ResMgr::Resource *scan=0;
+   Resource *scan=0;
    const char *closure;
    for(;;)
    {
@@ -2353,11 +2464,13 @@ void Http::LogErrorText()
    int size=conn->recv_buf->Size();
    if(size==0)
       return;
-   char *buf=string_alloca(size+1);
-   size=_Read(buf,size);
-   if(size<0)
+   Buffer tmpbuf;
+   size=_Read(&tmpbuf,size);
+   if(size<=0)
       return;
-   buf[size]=0;
+   tmpbuf.SpaceAdd(size);
+   const char *buf0=tmpbuf.Get();
+   char *buf=alloca_strdup(buf0);
    remove_tags(buf);
    for(char *line=strtok(buf,"\n"); line; line=strtok(0,"\n")) {
       rtrim(line);
@@ -2388,6 +2501,7 @@ check_end (const char *p)
     ++p;
   if (!*p
       || (p[0] == 'G' && p[1] == 'M' && p[2] == 'T')
+      || (p[0] == 'U' && p[1] == 'T' && p[2] == 'C')
       || ((p[0] == '+' || p[1] == '-') && ISDIGIT (p[1])))
     return 1;
   else
@@ -2469,6 +2583,27 @@ Http::atotm (const char *time_string)
    return ut;
 }
 
+bool Http::IsCompressed(const char *s)
+{
+   static const char *const values[] = {
+      "x-gzip", "gzip", "deflate", "compress", "x-compress", NULL
+   };
+   for(const char *const *v=values; *v; v++)
+      if(!strcmp(s,*v))
+	 return true;
+   return false;
+}
+
+bool Http::CompressedContentEncoding() const
+{
+   return content_encoding && IsCompressed(content_encoding);
+}
+bool Http::CompressedContentType() const
+{
+   static const char app[]="application/";
+   return entity_content_type && entity_content_type.begins_with(app)
+      && IsCompressed(entity_content_type+sizeof(app)-1);
+}
 
 #include "modconfig.h"
 #ifdef MODULE_PROTO_HTTP
