@@ -105,7 +105,7 @@ final:
 		     stats.mod_files,stats.mod_symlinks),
 	 tab,stats.mod_files,stats.mod_symlinks);
    if(stats.bytes)
-      s.appendf("%s%s\n",tab,CopyJob::FormatBytesTimeRate(stats.bytes,stats.time));
+      s.appendf("%s%s\n",tab,CopyJob::FormatBytesTimeRate(stats.bytes,transfer_time_elapsed));
    if(stats.del_dirs || stats.del_files || stats.del_symlinks)
       s.appendf(plural(flags&DELETE ?
 	       "%sRemoved: %d director$y|ies$, %d file$|s$, %d symlink$|s$\n"
@@ -178,6 +178,50 @@ void  MirrorJob::ShowRunStatus(const SMTaskRef<StatusLine>& s)
    }
 }
 
+xstring& MirrorJob::FormatShortStatus(xstring& s)
+{
+   if(bytes_to_transfer>0 && (!parent_mirror || parent_mirror->bytes_to_transfer!=bytes_to_transfer)) {
+      long long curr_bytes_transferred=GetBytesCount();
+      s.appendf("%s/%s (%d%%)",
+	 xhuman(curr_bytes_transferred),xhuman(bytes_to_transfer),
+	 percent(curr_bytes_transferred,bytes_to_transfer));
+      double rate=GetTransferRate();
+      if(rate>=1)
+	 s.append(' ').append(Speedometer::GetStrProper(rate));
+   }
+   return s;
+}
+
+void MirrorJob::TransferStarted(CopyJob *cp)
+{
+   if(transfer_count==0)
+      root_mirror->transfer_start_ts=now;
+   JobStarted(cp);
+}
+void MirrorJob::JobStarted(Job *j)
+{
+   AddWaiting(j);
+   transfer_count++;
+}
+void MirrorJob::TransferFinished(Job *j)
+{
+   long long bytes_count=j->GetBytesCount();
+   AddBytesTransferred(bytes_count);
+   stats.bytes+=bytes_count;
+   stats.time +=j->GetTimeSpent();
+   JobFinished(j);
+   if(transfer_count==0)
+      root_mirror->transfer_time_elapsed += now-root_mirror->transfer_start_ts;
+}
+void MirrorJob::JobFinished(Job *j)
+{
+   if(j->ExitCode()!=0)
+      stats.error_count++;
+   RemoveWaiting(j);
+   Delete(j);
+   transfer_count--;
+}
+
 void  MirrorJob::HandleFile(FileInfo *file)
 {
    int	 res;
@@ -315,8 +359,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    cp->SetDate(file->date);
 	 if(file->defined&file->SIZE)
 	    cp->SetSize(file->size);
-	 AddWaiting(cp);
-	 transfer_count++;
+	 TransferStarted(cp);
 	 cp->cmdline.vset("\\transfer ",file->name.get(),NULL);
 
 	 set_state(WAITING_FOR_TRANSFER);
@@ -546,10 +589,16 @@ void  MirrorJob::InitSets(const FileSet *source,const FileSet *dest)
    if(!(flags&DELETE))
       to_transfer->SubtractAny(to_rm_mismatched);
 
-   to_transfer->SortByPatternList(ResMgr::Query("mirror:order",0));
-   to_transfer->CountBytes(&bytes_to_transfer);
-   if(parent_mirror)
-      parent_mirror->AddBytesToTransfer(bytes_to_transfer);
+   const char *sort_by=ResMgr::Query("mirror:sort-by",0);
+   bool desc=strstr(sort_by,"-desc");
+   if(!strncmp(sort_by,"name",4))
+      to_transfer->SortByPatternList(ResMgr::Query("mirror:order",0));
+   else if(!strncmp(sort_by,"date",4))
+      to_transfer->Sort(FileSet::BYDATE);
+   else if(!strncmp(sort_by,"size",4))
+      to_transfer->Sort(FileSet::BYSIZE,false,true);
+   if(desc)
+      to_transfer->ReverseSort();
 
    int dir_count=0;
    to_transfer->Count(&dir_count,NULL,NULL,NULL);
@@ -609,7 +658,7 @@ void MirrorJob::HandleChdir(FileAccessRef& session, int &redirections)
    if(res==FA::OK)
       session->Close();
 }
-void MirrorJob::HandleListInfoCreation(const FileAccessRef& session,Ref<ListInfo>& list_info,const char *relative_dir)
+void MirrorJob::HandleListInfoCreation(const FileAccessRef& session,SMTaskRef<ListInfo>& list_info,const char *relative_dir)
 {
    if(state!=GETTING_LIST_INFO)
       return;
@@ -643,7 +692,7 @@ void MirrorJob::HandleListInfoCreation(const FileAccessRef& session,Ref<ListInfo
    list_info->Roll();
 }
 
-void MirrorJob::HandleListInfo(Ref<ListInfo>& list_info, Ref<FileSet>& set)
+void MirrorJob::HandleListInfo(SMTaskRef<ListInfo>& list_info, Ref<FileSet>& set)
 {
    if(!list_info)
       return;
@@ -808,10 +857,16 @@ int   MirrorJob::Do()
       stats.dirs++;
 
       InitSets(source_set,target_set);
+
+      to_transfer->CountBytes(&bytes_to_transfer);
+      if(parent_mirror)
+	 parent_mirror->AddBytesToTransfer(bytes_to_transfer);
+
       to_rm->Count(&stats.del_dirs,&stats.del_files,&stats.del_symlinks,&stats.del_files);
       to_rm->rewind();
       to_rm_mismatched->Count(&stats.del_dirs,&stats.del_files,&stats.del_symlinks,&stats.del_files);
       to_rm_mismatched->rewind();
+
       set_state(TARGET_REMOVE_OLD_FIRST);
       goto TARGET_REMOVE_OLD_FIRST_label;
 
@@ -823,15 +878,7 @@ int   MirrorJob::Do()
    case(WAITING_FOR_TRANSFER):
       while((j=FindDoneAwaitedJob())!=0)
       {
-	 long long bytes_count=j->GetBytesCount();
-	 AddBytesTransferred(bytes_count);
-	 stats.bytes+=bytes_count;
-	 stats.time +=j->GetTimeSpent();
-	 if(j->ExitCode()!=0)
-	    stats.error_count++;
-	 RemoveWaiting(j);
-	 Delete(j);
-	 transfer_count--;
+	 TransferFinished(j);
 	 m=MOVED;
       }
       if(max_error_count>0 && stats.error_count>=max_error_count)
@@ -841,6 +888,7 @@ int   MirrorJob::Do()
 	 file=to_transfer->curr();
       	 if(!file)
 	 {
+	    // go to the next step only when all transfers have finished
 	    if(waiting_num>0)
 	       break;
 	    if(FlagSet(DEPTH_FIRST))
@@ -876,11 +924,7 @@ int   MirrorJob::Do()
    TARGET_REMOVE_OLD_FIRST_label:
       while((j=FindDoneAwaitedJob())!=0)
       {
-	 if(j->ExitCode()!=0)
-	    stats.error_count++;
-	 RemoveWaiting(j);
-	 Delete(j);
-	 transfer_count--;
+	 JobFinished(j);
 	 m=MOVED;
       }
       if(max_error_count>0 && stats.error_count>=max_error_count)
@@ -940,8 +984,7 @@ int   MirrorJob::Do()
 	    args->seek(1);
 	    rmJob *j=new rmJob(target_session->Clone(),args);
 	    j->cmdline.set_allocated(args->Combine());
-	    AddWaiting(j);
-	    transfer_count++;
+	    JobStarted(j);
 	    if(file->defined&file->TYPE && file->filetype==file->DIRECTORY)
 	    {
 	       if(flags&NO_RECURSION)
@@ -977,9 +1020,7 @@ int   MirrorJob::Do()
    case(TARGET_CHMOD):
       while((j=FindDoneAwaitedJob())!=0)
       {
-	 RemoveWaiting(j);
-	 Delete(j);
-	 transfer_count--;
+	 JobFinished(j);
 	 m=MOVED;
       }
       if(max_error_count>0 && stats.error_count>=max_error_count)
@@ -1023,10 +1064,9 @@ int   MirrorJob::Do()
 	    a->seek(1);
 	    ChmodJob *cj=new ChmodJob(target_session->Clone(),
 				 file->mode&~mode_mask,a);
-	    AddWaiting(cj);
-	    transfer_count++;
 	    cj->cmdline.set_allocated(a->Combine());
 	    cj->BeQuiet(); // chmod is not supported on all servers; be quiet.
+	    JobStarted(cj);
 	    m=MOVED;
 	 }
       }
@@ -1049,9 +1089,7 @@ int   MirrorJob::Do()
    case(FINISHING):
       while((j=FindDoneAwaitedJob())!=0)
       {
-	 RemoveWaiting(j);
-	 Delete(j);
-	 transfer_count--;
+	 JobFinished(j);
 	 m=MOVED;
       }
       if(waiting_num>0)
@@ -1090,6 +1128,8 @@ int   MirrorJob::Do()
 	 PrintStatus(0,"");
 	 printf(_("Retrying mirror...\n"));
 	 stats.Reset();
+	 source_set=0;
+	 target_set=0;
 	 goto pre_GETTING_LIST_INFO;
       }
       /*fallthrough*/
@@ -1108,11 +1148,11 @@ MirrorJob::MirrorJob(MirrorJob *parent,
  :
    bytes_transferred(0), bytes_to_transfer(0),
    source_dir(new_source_dir), target_dir(new_target_dir),
-   root_transfer_count(0),
-   transfer_count(parent?parent->transfer_count:root_transfer_count)
+   transfer_time_elapsed(0), root_transfer_count(0),
+   transfer_count(parent?parent->transfer_count:root_transfer_count),
+   verbose_report(0),
+   parent_mirror(parent), root_mirror(parent?parent->root_mirror:this)
 {
-   verbose_report=0;
-   parent_mirror=parent;
 
    source_session=source;
    target_session=target;
