@@ -38,7 +38,6 @@ void NetAccess::Init()
    resolver=0;
    idle_timer.SetResource("net:idle",0);
    timeout_timer.SetResource("net:timeout",0);
-   max_retries=0;
    max_persist_retries=0;
    persist_retries=0;
    socket_buffer=0;
@@ -104,41 +103,28 @@ void NetAccess::Reconfig(const char *name)
       rate_limit->Reconfig(name,c);
 }
 
-int NetAccess::CheckHangup(const struct pollfd *pfd,int num)
+const char *NetAccess::CheckHangup(const struct pollfd *pfd,int num)
 {
    for(int i=0; i<num; i++)
    {
 #ifdef SO_ERROR
-      int   s_errno=0;
-      socklen_t len;
-
+      int s_errno=0;
       errno=0;
-
+      socklen_t len=sizeof(s_errno);
+      getsockopt(pfd[i].fd,SOL_SOCKET,SO_ERROR,(char*)&s_errno,&len);
 // Where does the error number go - to errno or to the pointer?
 // It seems that to errno, but if the pointer is NULL it dumps core.
 // (solaris 2.5)
 // It seems to be different on glibc 2.0 - check both errno and s_errno
-
-      len=sizeof(s_errno);
-      getsockopt(pfd[i].fd,SOL_SOCKET,SO_ERROR,(char*)&s_errno,&len);
-      if(errno==ENOTSOCK)
-	 return 0;
-      if(errno!=0 || s_errno!=0)
-      {
-	 LogError(0,_("Socket error (%s) - reconnecting"),
-				    strerror(errno?errno:s_errno));
-	 return 1;
-      }
+      if((errno!=0 || s_errno!=0) && errno!=ENOTSOCK)
+	 return strerror(errno?errno:s_errno);
 #endif /* SO_ERROR */
       if(pfd[i].revents&POLLERR)
-      {
-	 LogError(0,"POLLERR on fd %d",pfd[i].fd);
-	 return 1;
-      }
+	 return "POLLERR";
    } /* end for */
    return 0;
 }
-int NetAccess::Poll(int fd,int ev)
+int NetAccess::Poll(int fd,int ev,const char **err)
 {
    struct pollfd pfd;
    pfd.fd=fd;
@@ -147,7 +133,8 @@ int NetAccess::Poll(int fd,int ev)
    int res=poll(&pfd,1,0);
    if(res<1)
       return 0;
-   if(CheckHangup(&pfd,1))
+   *err=CheckHangup(&pfd,1);
+   if(*err)
       return -1;
    if(pfd.revents)
       timeout_timer.Reset();
@@ -241,7 +228,7 @@ void NetAccess::NextPeer()
    if(peer_curr>=peer.count())
       peer_curr=0;
    else
-      try_time=0;	// try next address immediately
+      DontSleep(); // try next address immediately
 }
 
 void NetAccess::ResetLocationData()
@@ -299,11 +286,8 @@ bool NetAccess::ReconnectAllowed()
       return true; // it will fault later - no need to wait.
    if(connection_limit>0 && connection_limit<=CountConnections())
       return false;
-   if(try_time==0)
+   if(reconnect_timer.Stopped())
       return true;
-   if(time_t(now) >= try_time+long(reconnect_interval_current))
-      return true;
-   TimeoutS(long(reconnect_interval_current)-(time_t(now)-try_time));
    return false;
 }
 
@@ -311,10 +295,12 @@ const char *NetAccess::DelayingMessage()
 {
    if(connection_limit>0 && connection_limit<=CountConnections())
       return _("Connection limit reached");
-   long remains=long(reconnect_interval_current)-(time_t(now)-try_time);
+   long remains=reconnect_timer.TimeLeft();
    if(remains<=0)
       return "";
    current->TimeoutS(1);
+   if(last_disconnect_cause && reconnect_timer.TimePassed()<5)
+      return last_disconnect_cause;
    return xstring::format("%s: %ld",_("Delaying before reconnect"),remains);
 }
 
@@ -340,7 +326,7 @@ bool NetAccess::CheckRetries()
       Fatal(_("max-retries exceeded"));
       return false;
    }
-   try_time=now;
+   reconnect_timer.Set(reconnect_interval_current);
    LogNote(10,"attempt number %d",retries);
    return true;
 }
