@@ -23,33 +23,46 @@
 #include "PollVec.h"
 #include "TimeDate.h"
 #include "Ref.h"
+#include "xarray.h"
+#include "xlist.h"
+#include "misc.h"
 
 class SMTask
 {
    virtual int Do() = 0;
 
    // all tasks list
-   SMTask *next;
-   static SMTask *chain;
+   static xlist_head<SMTask> all_tasks;
+   xlist<SMTask> all_tasks_node;
 
    // ready (not suspended) tasks list
-   SMTask *next_ready;
-   SMTask **prev_next_ready;
-   static SMTask *chain_ready;
-   void AddToReadyList();
-   void RemoveFromReadyList();
+   static xlist_head<SMTask> ready_tasks;
+   xlist<SMTask> ready_tasks_node;
+
+   // just created or resumed tasks
+   static xlist_head<SMTask> new_tasks;
+   xlist<SMTask> new_tasks_node;
+
+   // deleted and going to be destroyed tasks
+   static xlist_head<SMTask> deleted_tasks;
+   xlist<SMTask> deleted_tasks_node;
 
    static PollVec block;
-   static xarray<SMTask*> stack;
+   enum { SMTASK_MAX_DEPTH=64 };
+   static SMTask *stack[SMTASK_MAX_DEPTH];
+   static int stack_ptr;
 
    bool	 suspended;
    bool	 suspended_slave;
 
-protected:
    int	 running;
    int	 ref_count;
    bool	 deleting;
 
+   int ScheduleThis();
+   static int ScheduleNew();
+
+protected:
    enum
    {
       STALL=0,
@@ -58,14 +71,20 @@ protected:
    };
 
    // SuspendInternal and ResumeInternal usually suspend and resume slave tasks
-   virtual void SuspendInternal() { RemoveFromReadyList(); }
-   virtual void ResumeInternal() { AddToReadyList(); }
+   virtual void SuspendInternal() {}
+   virtual void ResumeInternal();
    virtual void PrepareToDie() {}  // it is called from Delete no matter of running and ref_count
 
+   bool Deleted() const { return deleting; }
+   virtual ~SMTask();
+
 public:
    static void Block(int fd,int mask) { block.AddFD(fd,mask); }
-   static void Timeout(int ms) { block.AddTimeout(ms); }
-   static void TimeoutS(int s) { Timeout(1000*s); }
+   static void TimeoutU(int us) { block.AddTimeoutU(us); }
+   static void Timeout(int ms) { TimeoutU(1000*ms); }
+   static void TimeoutS(int s) { TimeoutU(1000000*s); }
+   static bool Ready(int fd,int mask) { return block.FDReady(fd,mask); }
+   static void SetNotReady(int fd,int mask) { block.FDSetNotReady(fd,mask); }
 
    static TimeDate now;
    static void UpdateNow() { now.SetToCurrentTime(); }
@@ -87,7 +106,6 @@ public:
    virtual const char *GetLogContext() { return 0; }
 
    SMTask();
-   virtual ~SMTask();
 
    void DeleteLater();
    static void Delete(SMTask *);
@@ -111,7 +129,7 @@ public:
    static int TaskCount();
    static void PrintTasks();
    static bool NonFatalError(int err);
-   static bool TemporaryNetworkError(int err);
+   static bool TemporaryNetworkError(int err) { return temporary_network_error(err); }
 
    static void Cleanup();
 };
@@ -124,17 +142,32 @@ public:
    ~SMTaskInit();
 };
 
-template<class T> class SMTaskRef : public Ref<T>
+template<class T> class SMTaskRef
 {
    SMTaskRef<T>(const SMTaskRef<T>&);  // disable cloning
    void operator=(const SMTaskRef<T>&);   // and assignment
 
+protected:
+   T *ptr;
+
 public:
-   SMTaskRef() {}
-   SMTaskRef<T>(T *p) : Ref<T>(SMTask::MakeRef(p)) {}
-   ~SMTaskRef<T>() { SMTask::_DeleteRef(this->ptr); this->ptr=0; }
-   T *borrow() { if(this->ptr) this->ptr->DecRefCount(); return Ref<T>::borrow(); }
-   void operator=(T *p) { this->ptr=static_cast<T*>(SMTask::_SetRef(this->ptr,p)); }
+   SMTaskRef() { ptr=0; }
+   SMTaskRef<T>(T *p) : ptr(SMTask::MakeRef(p)) {}
+   ~SMTaskRef<T>() { SMTask::_DeleteRef(ptr); ptr=0; }
+   void operator=(T *p) { ptr=static_cast<T*>(SMTask::_SetRef(ptr,p)); }
+   operator const T*() const { return ptr; }
+   T *operator->() const { return ptr; }
+   T *borrow() { if(ptr) ptr->DecRefCount(); return replace_value(ptr,(T*)0); }
+   const T *get() const { return ptr; }
+   T *get_non_const() const { return ptr; }
+
+   template<class C> const SMTaskRef<C>& Cast() const
+      { void(static_cast<C*>(ptr)); return *(const SMTaskRef<C>*)this; }
+
+   static const SMTaskRef<T> null;
+
+   void _set(T *p) { ptr=p; }
+   void _clear() { ptr=0; }
    void unset() { *this=0; }
 };
 
