@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2013 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2014 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -53,20 +53,105 @@ public:
    int get_bit_length() const { return bit_length; }
    void set_bit_length(int b) { bit_length=b; set_length((b+7)/8); }
    void clear() { memset(buf,0,length()); }
+   void set_range(int from,int to,bool value);
 };
 
-struct TorrentPiece
+class TorrentPiece
 {
    unsigned sources_count;	    // how many peers have the piece
+   unsigned downloader_count;	    // how many downloaders of the piece are there
+   RefToArray<const TorrentPeer*> downloader; // which peers download the blocks
+   Ref<BitField> block_map;	    // which blocks are present.
 
-   BitField block_map;		    // which blocks are present
-   xarray<const TorrentPeer*> downloader; // which peers download the blocks
+public:
+   TorrentPiece() : sources_count(0), downloader_count(0) {}
+   ~TorrentPiece() {}
 
-   TorrentPiece(unsigned b)
-      : sources_count(0), block_map(b)
-      { downloader.allocate(b,0); }
+   unsigned get_sources_count() const { return sources_count; }
+   void add_sources_count(int diff) { sources_count+=diff; }
+   bool has_no_sources() const { return sources_count==0; }
 
-   bool has_a_downloader() const;
+   bool has_a_downloader() const { return downloader_count>0; }
+   void set_downloader(unsigned block,const TorrentPeer *o,const TorrentPeer *n,unsigned blk_count) {
+      if(!downloader) {
+	 if(o || !n)
+	    return;
+	 downloader=new const TorrentPeer*[blk_count]();
+      }
+      const TorrentPeer*& d=downloader[block];
+      if(d==o) {
+	 d=n;
+	 downloader_count+=(n!=0)-(o!=0);
+      }
+   }
+   void cleanup() {
+      if(downloader_count==0 && downloader)
+	 downloader=0;
+   }
+   const TorrentPeer *downloader_for(unsigned block) {
+      return downloader ? downloader[block] : 0;
+   }
+
+   void set_block_present(unsigned block,unsigned blk_count) {
+      if(!block_map)
+	 block_map=new BitField(blk_count);
+      block_map->set_bit(block,1);
+   }
+   void set_blocks_absent() {
+      block_map=0;
+   }
+   void free_block_map() {
+      block_map=0;
+   }
+   bool block_present(unsigned block) const {
+      return block_map && block_map->get_bit(block);
+   }
+   bool all_blocks_present(unsigned blk_count) const {
+      return block_map && block_map->has_all_set(0,blk_count);
+   }
+   bool any_blocks_present() const {
+      return block_map; // it's allocated when setting any bit
+   }
+};
+
+struct TorrentFile
+{
+   char *path;
+   off_t pos;
+   off_t length;
+   void set(const char *n,off_t p,off_t l) {
+      path=xstrdup(n);
+      pos=p;
+      length=l;
+   }
+   void unset() {
+      xfree(path); path=0;
+   }
+   bool contains_pos(off_t p) const {
+      return p>=pos && p<pos+length;
+   }
+};
+
+class TorrentFiles : public xarray<TorrentFile>
+{
+   static int pos_cmp(const TorrentFile *a, const TorrentFile *b) {
+      if(a->pos < b->pos)
+	 return -1;
+      if(a->pos > b->pos)
+	 return 1;
+      // we want zero-sized files to placed before non-zero ones.
+      if(a->length != b->length)
+	 return a->length < b->length ? -1 : 1;
+      return 0;
+   }
+public:
+   TorrentFile *file(int i) { return get_non_const()+i; }
+   TorrentFiles(const BeNode *f_node,const Torrent *t);
+   ~TorrentFiles() {
+      for(int i=0; i<length(); i++)
+	 file(i)->unset();
+   }
+   TorrentFile *FindByPosition(off_t p);
 };
 
 class TorrentListener : public SMTask, protected ProtoLog, protected Networker
@@ -99,6 +184,7 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    friend class TorrentPeer;
    friend class TorrentDispatcher;
    friend class TorrentListener;
+   friend class TorrentFiles;
    friend class DHT;
 
    bool shutting_down;
@@ -123,7 +209,7 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    static SMTaskRef<TorrentListener> listener_ipv6_udp;
    static SMTaskRef<DHT> dht_ipv6;
 #endif
-   static Ref<FDCache> fd_cache;
+   static SMTaskRef<FDCache> fd_cache;
    static Ref<TorrentBlackList> black_list;
 
    static const SMTaskRef<DHT>& GetDHT(int af)
@@ -154,10 +240,14 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
 
    xstring md_download;
    size_t metadata_size;
-   void StartMetadataDownload() { md_download.nset("",0); }
+   void FetchMetadataFromURL(const char *url);
+   void StartMetadataDownload();
    void MetadataDownloaded();
    void SetMetadata(const xstring& md);
    void ParseMagnet(const char *p);
+   const char *GetMetadataPath() const;
+   void SaveMetadata() const;
+   bool LoadMetadata(const char *path);
 
    xstring_c metainfo_url;
    SMTaskRef<FileCopy> metainfo_copy;
@@ -167,6 +257,7 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    xstring info_hash;
    const xstring *pieces;
    xstring name;
+   Ref<TorrentFiles> files;
 
    Ref<DirectedBuffer> recv_translate;
    Ref<DirectedBuffer> recv_translate_utf8;
@@ -210,16 +301,42 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    BeNode *Lookup(Ref<BeNode>& d,const char *name,BeNode::be_type_t type) { return Lookup(d->dict,name,type); }
 
    TaskRefArray<TorrentPeer> peers;
-   RefArray<TorrentPiece> piece_info;
    static int PeersCompareActivity(const SMTaskRef<TorrentPeer> *p1,const SMTaskRef<TorrentPeer> *p2);
    static int PeersCompareRecvRate(const SMTaskRef<TorrentPeer> *p1,const SMTaskRef<TorrentPeer> *p2);
    static int PeersCompareSendRate(const SMTaskRef<TorrentPeer> *p1,const SMTaskRef<TorrentPeer> *p2);
 
+   RefToArray<TorrentPiece> piece_info;
+   unsigned blocks_in_piece;
+   unsigned blocks_in_last_piece;
+   bool BlockPresent(unsigned piece,unsigned block) const {
+      return piece_info[piece].block_present(block);
+   }
+   bool AllBlocksPresent(unsigned piece) const {
+      return piece_info[piece].all_blocks_present(BlocksInPiece(piece));
+   }
+   bool AnyBlocksPresent(unsigned piece) const {
+      return piece_info[piece].any_blocks_present();
+   }
+   bool AllBlocksAbsent(unsigned piece) const {
+      return !AnyBlocksPresent(piece);
+   }
+   void SetBlocksAbsent(unsigned piece) {
+      piece_info[piece].set_blocks_absent();
+   }
+   void SetBlockPresent(unsigned piece,unsigned block) {
+      piece_info[piece].set_block_present(block,BlocksInPiece(piece));
+   }
+
+   void RebuildPiecesNeeded();
    Timer pieces_needed_rebuild_timer;
    xarray<unsigned> pieces_needed;
    static int PiecesNeededCmp(const unsigned *a,const unsigned *b);
    unsigned last_piece;
 
+   unsigned min_piece_sources;
+   unsigned avg_piece_sources;
+   unsigned pieces_available_pct;
+
    void SetPieceNotWanted(unsigned piece);
    void SetDownloader(unsigned piece,unsigned block,const TorrentPeer *o,const TorrentPeer *n);
 
@@ -256,6 +373,7 @@ class Torrent : public SMTask, protected ProtoLog, public ResClient
    Timer optimistic_unchoke_timer;
    Timer peers_scan_timer;
    Timer am_interested_timer;
+   Timer shutting_down_timer;
 
    Timer dht_announce_timer;
    int dht_announce_count;
@@ -286,6 +404,7 @@ public:
    static void ClassInit();
 
    Torrent(const char *mf,const char *cwd,const char *output_dir);
+   ~Torrent();
 
    int Do();
    int Done() const;
@@ -295,7 +414,7 @@ public:
    const Error *GetInvalidCause() const { return invalid_cause; }
 
    void Shutdown();
-   bool ShuttingDown() { return shutting_down; }
+   bool ShuttingDown() const { return shutting_down; }
    void PrepareToDie();
 
    bool CanAccept() const;
@@ -305,7 +424,7 @@ public:
    static void SHA1(const xstring& str,xstring& buf);
    void ValidatePiece(unsigned p);
    unsigned PieceLength(unsigned p) const { return p==total_pieces-1 ? last_piece_length : piece_length; }
-   unsigned BlocksInPiece(unsigned p) const { return (PieceLength(p)+BLOCK_SIZE-1)/BLOCK_SIZE; }
+   unsigned BlocksInPiece(unsigned p) const { return p==total_pieces-1 ? blocks_in_last_piece : blocks_in_piece; }
 
    const TaskRefArray<TorrentPeer>& GetPeers() const { return peers; }
    void AddPeer(TorrentPeer *);
@@ -320,9 +439,14 @@ public:
    bool Complete() const { return complete; }
    bool Private() const { return is_private; }
    double GetRatio() const;
+   void CalcPiecesStats();
+   unsigned MinPieceSources() const { return min_piece_sources; }
+   double AvgPieceSources() const { return avg_piece_sources/256.; }
+   unsigned PiecesAvailablePct() const { return pieces_available_pct; }
    unsigned long long TotalLength() const { return total_length; }
    unsigned PieceLength() const { return piece_length; }
    const char *GetName() const { return name?name.get():metainfo_url.get(); }
+   bool IsDownloading() const { return HasMetadata() && !IsValidating() && !Complete() && !ShuttingDown(); }
 
    void Reconfig(const char *name);
    const char *GetLogContext() { return GetName(); }
@@ -346,7 +470,7 @@ public:
    unsigned long long GetTotalLeft() { return total_left; }
 
    const TaskRefArray<TorrentTracker>& Trackers() { return trackers; }
-   bool HasMetadata() { return metadata!=0; }
+   bool HasMetadata() const { return metadata!=0; }
    void DisconnectPeers();
 
    static void BootstrapDHT(const char *n) {
@@ -418,6 +542,8 @@ class TorrentPeer : public SMTask, protected ProtoLog, public Networker
    bool connected;
    bool passive;
 
+   xstring_c last_dc;
+
    Timer timeout_timer;
    Timer retry_timer;
    Timer keepalive_timer;
@@ -691,7 +817,7 @@ private:
    void SendHandshake();
    unpack_status_t RecvHandshake();
    void SendExtensions();
-   void Disconnect();
+   void Disconnect(const char *dc=0);
    int SendDataRequests(unsigned p);
    void SendDataRequests();
    void Have(unsigned p);
@@ -755,7 +881,7 @@ public:
 
 class TorrentBlackList
 {
-   xmap<Timer*> bl;
+   xmap_p<Timer> bl;
    void check_expire();
 public:
    bool Listed(const sockaddr_u &a);
@@ -768,11 +894,12 @@ class TorrentDispatcher : public SMTask, protected ProtoLog
    const sockaddr_u addr;
    SMTaskRef<IOBuffer> recv_buf;
    Timer timeout_timer;
+   xstring_c peer_name;
 public:
    TorrentDispatcher(int s,const sockaddr_u *a);
    ~TorrentDispatcher();
    int Do();
-   const char *GetLogContext() { return "torrent"; }
+   const char *GetLogContext() { return peer_name; }
 };
 
 #include "Job.h"
