@@ -191,6 +191,31 @@ void Buffer::Empty()
       save=true;
 }
 
+// move data from other buffer, prepare for SpaceAdd.
+int Buffer::MoveDataHere(Buffer *o,int max_len)
+{
+   const char *b;
+   int size;
+   o->Get(&b,&size);
+   if(size>max_len)
+      size=max_len;
+   if(size>0) {
+      if(size>=64 && Size()==0 && o->Size()==size && !save && !o->save) {
+	 // optimization by swapping buffers
+	 buffer.swap(o->buffer);
+	 buffer_ptr=replace_value(o->buffer_ptr,buffer_ptr);
+	 buffer.set_length_no_z(buffer_ptr);
+	 o->pos+=size;
+      } else {
+	 memcpy(GetSpace(size),b,size);
+	 o->Skip(size);
+      }
+   }
+   if(o->Eof() && o->Size()==0)
+      PutEOF();
+   return size;
+}
+
 Buffer::Buffer()
 {
    saved_errno=0;
@@ -381,6 +406,16 @@ void DirectedBuffer::PutTranslated(const char *buf,int size)
    else
       Buffer::Put(buf,size);
 }
+int DirectedBuffer::MoveDataHere(Buffer *buf,int size)
+{
+   if(size>buf->Size())
+      size=buf->Size();
+   if(mode==PUT && translator)
+      translator->PutTranslated(this,buf->Get(),size);
+   else
+      return Buffer::MoveDataHere(buf,size);
+   return size;
+}
 
 void DirectedBuffer::EmbraceNewData(int len)
 {
@@ -400,7 +435,8 @@ void DirectedBuffer::EmbraceNewData(int len)
 
 
 IOBuffer::IOBuffer(dir_t m)
-   : DirectedBuffer(m), event_time(now), max_buf(0)
+   : DirectedBuffer(m), event_time(now),
+     max_buf(0), get_size(GET_BUFSIZE)
 {
 }
 IOBuffer::~IOBuffer()
@@ -430,6 +466,18 @@ void IOBuffer::Put(const char *buf)
    Put(buf,strlen(buf));
 }
 
+int IOBuffer::TuneGetSize(int res)
+{
+   if(res>0)
+   {
+      // buffer size tuning depending on data rate
+      const int max_get_size=(max_buf?max_buf:0x100000);
+      if(res>get_size/2 && Size()+get_size*2<=max_get_size)
+	 get_size*=2;
+   }
+   return res;
+}
+
 int IOBuffer::Do()
 {
    if(Done() || Error())
@@ -453,7 +501,7 @@ int IOBuffer::Do()
    case GET:
       if(eof)
 	 return STALL;
-      res=Get_LL(GET_BUFSIZE);
+      res=TuneGetSize(Get_LL(get_size));
       if(res>0)
       {
 	 EmbraceNewData(res);
@@ -504,15 +552,13 @@ int IOBufferStacked::Do()
       {
 	 buffer_ptr+=res;
 	 m=MOVED;
-	 down->Do();
       }
       break;
 
    case GET:
-      m|=down->Do();
       if(eof)
 	 return m;
-      res=Get_LL(GET_BUFSIZE);
+      res=Get_LL(/*unused*/0);
       if(res>0)
       {
 	 EmbraceNewData(res);
@@ -549,20 +595,7 @@ int IOBufferStacked::Get_LL(int)
       return 0;
    }
    down->ResumeSlave();
-
-   const char *b;
-   int size;
-   down->Get(&b,&size);
-
-   if(!b)
-   {
-      eof=true;
-      return 0;
-   }
-
-   memcpy(GetSpace(size),b,size);
-   down->Skip(size);
-   return size;
+   return MoveDataHere(down,down->Size());
 }
 
 bool IOBufferStacked::Done()
@@ -656,12 +689,19 @@ int IOBufferFDStream::Get_LL(int size)
       return 0;
    }
 
+   if(!Ready(fd,POLLIN))
+   {
+      Block(fd,POLLIN);
+      return 0;
+   }
+
    res=read(fd,GetSpace(size),size);
    if(res==-1)
    {
       saved_errno=errno;
       if(E_RETRY(saved_errno))
       {
+	 SetNotReady(fd,POLLIN);
 	 Block(fd,POLLIN);
 	 return 0;
       }
@@ -712,7 +752,7 @@ int IOBufferFileAccess::Get_LL(int size)
 
    int res=0;
 
-   res=session->Read(GetSpace(size),size);
+   res=session->Read(this,size);
    if(res<0)
    {
       if(res==FA::DO_AGAIN)
