@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2013 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2014 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -92,6 +92,8 @@ CMD(at);    CMD(find);   CMD(command); CMD(module);
 CMD(lpwd);  CMD(glob);	 CMD(chmod);   CMD(queue);
 CMD(repeat);CMD(get1);   CMD(tasks);   CMD(torrent);
 
+CMD(empty); CMD(notempty); CMD(true); CMD(false);
+
 #ifdef MODULE_CMD_MIRROR
 # define cmd_mirror 0
 #endif
@@ -211,9 +213,12 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	   )},
    {"connect", cmd_open,   0,"open"},
    {"command", cmd_command},
-   {"debug",   cmd_debug,  N_("debug [<level>|off] [-o <file>]"),
+   {"debug",   cmd_debug,  N_("debug [OPTS] [<level>|off]"),
 	 N_("Set debug level to given value or turn debug off completely.\n"
-	 " -o <file>  redirect debug output to the file.\n")},
+	 " -o <file>  redirect debug output to the file\n"
+	 " -c  show message context\n"
+	 " -p  show PID\n"
+	 " -t  show timestamps\n")},
    {"du",      cmd_du,  N_("du [options] <dirs>"),
 	 N_("Summarize disk usage.\n"
 	 " -a, --all             write counts for all files, not just directories\n"
@@ -331,9 +336,10 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 "If the second directory is omitted, basename of first directory is used.\n"
 	 "If both directories are omitted, current local and remote directories are used.\n"
 	 )},
-   {"mkdir",   cmd_mkdir,  N_("mkdir [-p] <dirs>"),
+   {"mkdir",   cmd_mkdir,  N_("mkdir [OPTS] <dirs>"),
 	 N_("Make remote directories\n"
-	 " -p  make all levels of path\n")},
+	 " -p  make all levels of path\n"
+	 " -f  be quiet, suppress messages\n")},
    {"module",  cmd_module, N_("module name [args]"),
 	 N_("Load module (shared object). The module should contain function\n"
 	 "   void module_init(int argc,const char *const *argv)\n"
@@ -485,6 +491,10 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 N_("Same as more, but filter each file through bzcat\n")},
 
    {".tasks",  cmd_tasks,  0,0},
+   {".empty",  cmd_empty,  0,0},
+   {".notempty",cmd_notempty,0,0},
+   {".true",   cmd_true,   0,0},
+   {".false",  cmd_false,  0,0},
 };
 const int CmdExec::static_cmd_table_length=sizeof(static_cmd_table)/sizeof(static_cmd_table[0]);
 
@@ -628,7 +638,7 @@ Job *CmdExec::builtin_cd()
    int cache_is_dir=FileAccess::cache->IsDirectory(session,dir);
    if(cache_is_dir==1)
    {
-      if(is_file && dir_needs_slash)
+      if(is_file && dir_needs_slash && last_char(dir)!='/')
 	 dir=xstring::get_tmp(dir).append('/');
       is_file=false;
    }
@@ -674,6 +684,11 @@ Job *CmdExec::builtin_exit()
 	 if(top)
 	    exec=top.get_non_const();
       }
+      else if(!strcmp(a,"parent"))
+      {
+	 if(parent_exec)
+	    exec=parent_exec;
+      }
       else if(!strcmp(a,"kill"))
       {
 	 kill=true;
@@ -687,7 +702,7 @@ Job *CmdExec::builtin_exit()
    }
    // Note: one job is this CmdExec.
    if(!bg && exec->top_level
-   && !ResMgr::QueryBool("cmd:move-background",0) && NumberOfJobs()>0)
+   && !ResMgr::QueryBool("cmd:move-background",0) && NumberOfChildrenJobs()>0)
    {
       eprintf(_(
 	 "There are running jobs and `cmd:move-background' is not set.\n"
@@ -695,13 +710,15 @@ Job *CmdExec::builtin_exit()
       ));
       return 0;
    }
-   if(!detach && Job::NumberOfJobs()==0)
+   if(!detach && Job::NumberOfChildrenJobs()==0)
       detach=true;
    if(kill)
       Job::KillAll();
-   if(detach)
+   if(detach) {
+      for(CmdExec *e=this; e!=exec; e=e->parent_exec)
+	 e->Exit(code);
       exec->Exit(code);
-   else {
+   } else {
       int loc=0;
       exec->SetAutoTerminateInBackground(true);
       eprintf(_(
@@ -733,6 +750,7 @@ void CmdExec::Exit(int code)
 {
    while(feeder)
       RemoveFeeder();
+   cmd_buf.Empty();
    if(interactive)
    {
       ListDoneJobs();
@@ -827,13 +845,15 @@ Job *CmdExec::builtin_open()
 
    enum {
       OPT_USER,
-      OPT_PASSWORD
+      OPT_PASSWORD,
+      OPT_ENV_PASSWORD,
    };
    static struct option open_options[]=
    {
       {"port",required_argument,0,'p'},
       {"user",required_argument,0,OPT_USER},
       {"password",required_argument,0,OPT_PASSWORD},
+      {"env-password",no_argument,0,OPT_ENV_PASSWORD},
       {"execute",required_argument,0,'e'},
       {"debug",optional_argument,0,'d'},
       {"no-bookmark",no_argument,0,'B'},
@@ -873,6 +893,9 @@ Job *CmdExec::builtin_open()
       case(OPT_PASSWORD):
 	 pass=optarg;
 	 break;
+      case(OPT_ENV_PASSWORD):
+	 pass=getenv("LFTP_PASSWORD");
+	 break;
       case('d'):
 	 debug=true;
 	 break;
@@ -1103,8 +1126,16 @@ Job *CmdExec::builtin_glob()
    const char *op=args->a0();
    int opt;
    GlobURL::type_select glob_type=GlobURL::FILES_ONLY;
+   const char *cmd=0;
 
-   while((opt=args->getopt("+adf"))!=EOF)
+   static struct option glob_options[]=
+   {
+      {"exist",no_argument,0,'e'},
+      {"not-exist",no_argument,0,'E'},
+      {0}
+   };
+
+   while((opt=args->getopt_long("+adf",glob_options))!=EOF)
    {
       switch(opt)
       {
@@ -1117,6 +1148,12 @@ Job *CmdExec::builtin_glob()
       case('f'):
 	 glob_type=GlobURL::FILES_ONLY;
 	 break;
+      case('e'):
+	 cmd=".notempty";
+	 break;
+      case('E'):
+	 cmd=".empty";
+	 break;
       case('?'):
 	 eprintf(_("Try `help %s' for more information.\n"),op);
 	 return 0;
@@ -1124,6 +1161,8 @@ Job *CmdExec::builtin_glob()
    }
    while(args->getindex()>1)
       args->delarg(1);	   // remove options.
+   if(cmd)
+      args->insarg(1,cmd);
    if(args->count()<2)
    {
       eprintf(_("Usage: %s [OPTS] command args...\n"),op);
@@ -2335,7 +2374,7 @@ CMD(wait)
       eprintf(_("%s: wait loop detected\n"),op);
       return 0;
    }
-   j->parent=0;
+   j->SetParent(0);
    j->Bg();
    return j;
 }
@@ -2584,7 +2623,7 @@ CMD(help)
 CMD(ver)
 {
    printf(
-      _("LFTP | Version %s | Copyright (c) 1996-%d Alexander V. Lukyanov\n"),VERSION,2013);
+      _("LFTP | Version %s | Copyright (c) 1996-%d Alexander V. Lukyanov\n"),VERSION,2014);
    printf("\n");
    printf(
       _("LFTP is free software: you can redistribute it and/or modify\n"
@@ -3331,6 +3370,27 @@ CMD(tasks)
    return 0;
 }
 
+CMD(empty)
+{
+   exit_code=(args->count()>1 ? 1 : 0);
+   return 0;
+}
+CMD(notempty)
+{
+   exit_code=(args->count()>1 ? 0 : 1);
+   return 0;
+}
+CMD(true)
+{
+   exit_code=0;
+   return 0;
+}
+CMD(false)
+{
+   exit_code=1;
+   return 0;
+}
+
 CMD(eval)
 {
    int opt;
