@@ -17,8 +17,6 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-/* $Id: ftpclass.h,v 1.166 2011/03/03 20:03:11 lav Exp $ */
-
 #ifndef FTPCLASS_H
 #define FTPCLASS_H
 
@@ -175,7 +173,7 @@ class Ftp : public NetAccess
       void SuspendInternal()
       {
 	 if(control_send) control_send->SuspendSlave();
-	 if(control_recv) control_recv->SuspendSlave();
+	 if(control_recv && !data_iobuf) control_recv->SuspendSlave();
 	 if(data_iobuf)	data_iobuf->SuspendSlave();
       }
       void ResumeInternal()
@@ -329,6 +327,7 @@ private:
    void	SendSiteIdle();
    void	SendAcct();
    void	SendSiteGroup();
+   void	SendSiteCommands();
    void	SendUTimeRequest();
    void SendAuth(const char *auth);
    void TuneConnectionAfterFEAT();
@@ -415,12 +414,16 @@ private:
 
    bool use_telnet_iac;
 
+   int max_buf;
+
    const char *get_protect_res();
    const char *encode_eprt(const sockaddr_u *);
 
    typedef FileInfo *(*FtpLineParser)(char *line,int *err,const char *tz);
    static FtpLineParser line_parsers[];
 
+   int CanRead();
+
 protected:
    ~Ftp();
 
@@ -457,7 +460,7 @@ public:
       return GetConnectLevel()!=CL_NOT_CONNECTED;
    }
 
-   int   Read(void *buf,int size);
+   int   Read(Buffer *buf,int size);
    int   Write(const void *buf,int size);
    int   Buffered();
    void  Close();
@@ -469,7 +472,7 @@ public:
    int	 StoreStatus();
 
    int	 Do();
-   void  Disconnect();
+   void  DisconnectLL();
    void  DisconnectNow();
 
    void	 SetFlag(int flag,bool val);
@@ -507,7 +510,7 @@ public:
 	 copy_protect=prot;
 	 copy_ssl_connect=sscn;
 	 retries=rnum;
-	 try_time=tt;
+	 SetTryTime(tt);
       }
    bool SetCopyAddress(const Ftp *o)
       {
