@@ -42,6 +42,7 @@
 
 CDECL_BEGIN
 #include "regex.h"
+#include "human.h"
 CDECL_END
 #include "misc.h"
 #include "ProcWait.h"
@@ -566,7 +567,7 @@ char *Subst(const char *txt, const subst_t *s)
       {
 	 ch=*txt++;
 	 if(ch >= '0' && ch < '8') {
-	    unsigned len;
+	    int len;
 	    unsigned code;
 	    txt--;
 	    if(sscanf(txt,"%3o%n",&code,&len)!=1)
@@ -740,6 +741,8 @@ bool temporary_network_error(int err)
 {
    switch(err)
    {
+   case(EPIPE):
+   case(EIO):
    case(ETIMEDOUT):
 #ifdef ECONNRESET
    case(ECONNRESET):
@@ -989,3 +992,9 @@ const char *get_nodename()
       return u.nodename;
    return "NODE";
 }
+
+const char *xhuman(long long n)
+{
+   char *buf=xstring::tmp_buf(LONGEST_HUMAN_READABLE + 1);
+   return human_readable(n, buf, human_autoscale|human_SI, 1, 1);
+}
