@@ -57,13 +57,39 @@ static const char *FtpProxyValidate(xstring_c *p)
    return 0;
 }
 
+static const char *SetValidate(xstring_c& s,const char *const *set,const char *name)
+{
+   const char *const *scan;
+   for(scan=set; *scan; scan++)
+      if(s.eq(*scan))
+	 return 0;
+
+   xstring &j=xstring::get_tmp();
+   if(name)
+      j.setf(_("%s must be one of: "),name);
+   else
+      j.set(_("must be one of: "));
+   bool had_empty=false;
+   for(scan=set; *scan; scan++) {
+      if(!**scan) {
+	 had_empty=true;
+	 continue;
+      }
+      if(scan>set)
+	 j.append(", ");
+      j.append(*scan);
+   }
+   if(had_empty)
+      j.append(_(", or empty"));
+   return j;
+}
+
 static const char *FtpProxyAuthTypeValidate(xstring_c *s)
 {
-   if(s->ne("user") && s->ne("joined") && s->ne("joined-acct")
-   && s->ne("open") && s->ne("proxy-user@host"))
-      // for translator: `user', `joined', `joined-acct', `open' are literals.
-      return _("ftp:proxy-auth-type must be one of: user, joined, joined-acct, open, proxy-user@host");
-   return 0;
+   static const char *const valid_set[]={
+      "user", "joined", "joined-acct", "open", "proxy-user@host", 0
+   };
+   return SetValidate(*s,valid_set,"ftp:proxy-auth-type");
 }
 
 static const char *HttpProxyValidate(xstring_c *p)
@@ -130,6 +156,14 @@ const char *OrderValidate(xstring_c *s)
    return 0;
 }
 
+static const char *SortByValidate(xstring_c *s)
+{
+   static const char * const valid_set[]={
+      "name", "name-desc", "size", "size-desc", "date", "date-desc", 0
+   };
+   return SetValidate(*s,valid_set,"mirror:order-by");
+}
+
 #if USE_SSL
 static
 const char *AuthArgValidate(xstring_c *s)
@@ -137,30 +171,21 @@ const char *AuthArgValidate(xstring_c *s)
    for(char *i=s->get_non_const(); *i; i++)
       *i=to_ascii_upper(*i);
 
-   if(strcmp(*s,"SSL")
-   && strcmp(*s,"TLS")
-   && strcmp(*s,"TLS-P")
-   && strcmp(*s,"TLS-C"))
-      return _("ftp:ssl-auth must be one of: SSL, TLS, TLS-P, TLS-C");
-
-   return 0;
+   const char *const valid_set[]={
+      "SSL", "TLS", "TLS-P", "TLS-C", 0
+   };
+   return SetValidate(*s,valid_set,"ftp:ssl-auth");
 }
 static
 const char *ProtValidate(xstring_c *s)
 {
-   if(!**s)
-      return 0;
-
    for(char *i=s->get_non_const(); *i; i++)
       *i=to_ascii_upper(*i);
 
-   if(strcmp(*s,"P")
-   && strcmp(*s,"C")
-   && strcmp(*s,"S")
-   && strcmp(*s,"E"))
-      return _("must be one of: C, S, E, P, or empty");
-
-   return 0;
+   const char *const valid_set[]={
+      "C", "S", "E", "P", "", 0
+   };
+   return SetValidate(*s,valid_set,"ftps:initial-prot");
 }
 #endif
 
@@ -173,7 +198,7 @@ static ResType lftp_vars[] = {
    {"ftp:auto-passive-mode",	 "yes",   ResMgr::BoolValidate,0},
    {"ftp:bind-data-socket",	 "yes",   ResMgr::BoolValidate,0},
    {"ftp:charset",		 "",	  ResMgr::CharsetValidate,0},
-   {"ftp:client",		 PACKAGE"/"VERSION,0,0},
+   {"ftp:client",		 PACKAGE "/" VERSION,0,0},
    {"ftp:device-prefix",	 "no",	  ResMgr::BoolValidate,0},
    {"ftp:fix-pasv-address",	 "yes",   ResMgr::BoolValidate,0},
    {"ftp:ignore-pasv-address",	 "no",	  ResMgr::BoolValidate,0},
@@ -181,6 +206,7 @@ static ResType lftp_vars[] = {
    {"ftp:fxp-passive-source",	 "no",	  ResMgr::BoolValidate,ResMgr::NoClosure},
    {"ftp:fxp-passive-sscn",	 "yes",   ResMgr::BoolValidate,ResMgr::NoClosure},
    {"ftp:home",			 "",	  0,0},
+   {"ftp:site"			 "",	  0,0},
    {"ftp:site-group",		 "",	  0,0},
    {"ftp:lang",			 "",	  0,0},
    {"ftp:list-empty-ok",	 "no",	  0,0},
@@ -248,6 +274,7 @@ static ResType lftp_vars[] = {
    {"ftp:retry-530-anonymous",	 "Login incorrect",ResMgr::ERegExpValidate,0},
    {"hftp:cache",		 "yes",   ResMgr::BoolValidate,0},
    {"hftp:cache-control",	 "",	  0,0},
+   {"hftp:decode",		 "no",	  ResMgr::BoolValidate,0},
    {"hftp:proxy",		 "",	  HttpProxyValidate,0},
    {"hftp:use-authorization",	 "yes",   ResMgr::BoolValidate,0},
    {"hftp:use-head",		 "yes",   ResMgr::BoolValidate,0},
@@ -258,14 +285,16 @@ static ResType lftp_vars[] = {
    {"http:accept",		 "*/*",   0,0},
    {"http:accept-language",	 "",	  0,0},
    {"http:accept-charset",	 "",	  0,0},
+   {"http:accept-encoding",	 "",	  0,0},
    {"http:authorization",	 "",	  0,0},
    {"http:cache",		 "yes",   ResMgr::BoolValidate,0},
    {"http:cache-control",	 "",	  0,0},
+   {"http:decode",		 "yes",	  ResMgr::BoolValidate,0},
    {"http:proxy",		 "",	  HttpProxyValidate,0},
    {"http:use-mkcol",		 "yes",   ResMgr::BoolValidate,0},
    {"http:use-propfind",	 "no",    ResMgr::BoolValidate,0},
    {"http:use-allprop",		 "no",	  ResMgr::BoolValidate,0},
-   {"http:user-agent",		 PACKAGE"/"VERSION,0,0},
+   {"http:user-agent",		 PACKAGE "/" VERSION,0,0},
    {"http:cookie",		 "",	  0,0},
    {"http:set-cookies",		 "no",	  0,0},
    {"http:post-content-type",	 "application/x-www-form-urlencoded",0,0},
@@ -296,6 +325,7 @@ static ResType lftp_vars[] = {
    {"net:connection-limit",	 "0",	  ResMgr::UNumberValidate,0},
    {"net:connection-takeover",	 "yes",   ResMgr::BoolValidate,0},
 
+   {"mirror:sort-by",		 "name",  SortByValidate,ResMgr::NoClosure},
    {"mirror:order",		 "*.sfv *.sig *.md5* *.sum * */", 0,ResMgr::NoClosure},
    {"mirror:parallel-directories", "yes", ResMgr::BoolValidate,ResMgr::NoClosure},
    {"mirror:parallel-transfer-count", "1",ResMgr::UNumberValidate,ResMgr::NoClosure},
@@ -347,6 +377,7 @@ static ResType lftp_vars[] = {
    {"xfer:log",			 "yes",	  ResMgr::BoolValidate,ResMgr::NoClosure},
    {"xfer:auto-rename",		 "no",	  ResMgr::BoolValidate,ResMgr::NoClosure},
    {"xfer:log-file",              "",     ResMgr::FileCreatable,ResMgr::NoClosure},
+   {"xfer:max-log-size",	 "1M",	  ResMgr::UNumberValidate,ResMgr::NoClosure},
 
 #if USE_SSL
    {"ssl:ca-file",		 "",	  ResMgr::FileReadable,ResMgr::NoClosure},
@@ -369,13 +400,13 @@ static ResDecls lftp_vars_register(lftp_vars);
 # include <langinfo.h>
 #endif
 
-bool ResMgr::class_inited;
-void ResMgr::ClassInit()
+bool ResType::class_inited;
+void ResType::ClassInit()
 {
    if(class_inited)
       return;
    class_inited=true;
-   for(ResType *scan=type_chain; scan; scan=scan->next)
+   for(ResType *scan=types_by_name->each_begin(); scan; scan=types_by_name->each_next())
    {
       if(scan->defvalue && scan->val_valid)
       {
