@@ -24,37 +24,6 @@
 #include "trio.h"
 #include "c-ctype.h"
 
-int xstrcmp(const char *s1,const char *s2)
-{
-   if(s1==s2)
-      return 0;
-   if(s1==0 || s2==0)
-      return 1;
-   return strcmp(s1,s2);
-}
-int xstrncmp(const char *s1,const char *s2,size_t len)
-{
-   if(s1==s2 || len==0)
-      return 0;
-   if(s1==0 || s2==0)
-      return 1;
-   return strncmp(s1,s2,len);
-}
-int xstrcasecmp(const char *s1,const char *s2)
-{
-   if(s1==s2)
-      return 0;
-   if(s1==0 || s2==0)
-      return 1;
-   return strcasecmp(s1,s2);
-}
-size_t xstrlen(const char *s)
-{
-   if(s==0)
-      return 0;
-   return strlen(s);
-}
-
 void xstring::get_space(size_t s)
 {
    get_space2(s,32);
@@ -65,10 +34,14 @@ void xstring::get_space2(size_t s,size_t g)
       buf=(char*)xmalloc(size=s+1);
    else if(size<s+1)
       buf=(char*)xrealloc(buf,size=(s|(g-1))+1);
-   else if(size>=g*8 && s+1<=size/2)
-      buf=(char*)xrealloc(buf,size/=2);
    buf[s]=0;
 }
+void xstring::shrink_space()
+{
+   enum { MIN_SIZE=128 };
+   if(buf && size>MIN_SIZE)
+      buf=(char*)xrealloc(buf,((len+1)|(MIN_SIZE-1))+1);
+}
 char *xstring::add_space(size_t s)
 {
    if(size<=len+s)
@@ -147,6 +120,12 @@ xstring& xstring::move_here(xstring& o)
    buf=o.buf; o.buf=0;
    return *this;
 }
+void xstring::swap(xstring& o)
+{
+   buf=replace_value(o.buf,buf);
+   size=replace_value(o.size,size);
+   len=replace_value(o.len,len);
+}
 
 xstring& xstring::append(const char *s,size_t s_len)
 {
@@ -361,6 +340,11 @@ xstring& xstring::get_tmp()
 {
    static xstring revolver[16];
    static int i;
+
+   enum { MAX_REVOLVER_SIZE=0x1000 };
+   if(revolver[i].size>MAX_REVOLVER_SIZE)
+      revolver[i].shrink_space();
+
    int next=(i+1)&15;
    xstring& tmp=revolver[i];
    // keep the oldest tmp clear to trigger NULL dereference
